package org.cis1200;

import org.junit.jupiter.api.*;
import static org.junit.jupiter.api.Assertions.*;

import java.util.Collection;

public class ServerModelTest {
    private ServerModel model;

    /**
     * Before each test, we initialize model to be
     * a new ServerModel (with all new, empty state).
     * The BeforeEach annotation makes it so that the
     * following method will get run before each unit
     * test automatically.
     */
    @BeforeEach
    public void setUp() {
        // We initialize a fresh ServerModel for each test
        model = new ServerModel();
    }

    /**
     * Here is an example test that checks the functionality of your
     * changeNickname error handling. Each line has commentary directly above
     * it which you can use as a framework for the remainder of your tests.
     */
    @Test
    public void testInvalidNickname() {
        // A user must be registered before their nickname can be changed,
        // so we first register a user with an arbitrarily chosen id of 0.
        model.registerUser(0);

        // We manually create a Command that appropriately tests the case
        // we are checking. In this case, we create a NicknameCommand whose
        // new Nickname is invalid.
        Command command = new NicknameCommand(0, "!nv@l!d!");

        // To test how the server processes this command, we need to do two things:
        // 1. we need to make sure that the set of responses that it produces matches
        // our expectations
        // 2. we need to make sure that the command updates the state of the server
        // in an appropriate way.

        // We ask the server to process the command using the updateServerModel method.
        // For the Nickname command, dynamic dispatch invokes the changeNickname method
        // in the ServerModel class.
        ResponseSet actual = command.updateServerModel(model);

        // We manually create the *expected* ResponseSet containing a single
        // response, constructed by the error static method. This error message
        // says that the NicknameCommand above contains an invalid name.
        ResponseSet expected = ResponseSet.singleMessage(
                Response.error(
                        command, ErrorCode.INVALID_NAME
                )
        );

        // The first assertEquals tests whether the changeNickname method returns
        // the expected ResponseSet.
        assertEquals(expected, actual, "error response for invalid nickname");

        // We also want to test whether the state has been correctly
        // changed.In this case, the state that would be affected is
        // the user's Collection.
        Collection<String> users = model.getRegisteredUsers();

        // We now check to see if our command updated the state
        // appropriately. In this case, we first ensure that no
        // additional users have been added.
        assertEquals(1, users.size(), "Number of registered users");

        // We then check if the username was updated to an invalid value
        // (it should not have been).
        assertTrue(users.contains("User0"), "Old nickname still registered");

        // Finally, we check that the id 0 is still associated with the old,
        // unchanged nickname.
        assertEquals(
                "User0", model.getNickname(0),
                "User with id 0 nickname unchanged"
        );
    }

    /**
     * Here's a simple test modified from testRegisterSingleUser
     * in Task3Test which demonstrates how you can use the constructor
     * and the different methods (addMessage, singleMessage)
     * of the ResponseSet class.
     */
    @Test
    public void testRegisterSingleUserResponseSet() {
        // test: does model.registerUser(0) return what we expect?
        // We create an expected ResponseSet by using the singleMessage
        // method in the ResponseSet class.
        ResponseSet expected = ResponseSet.singleMessage(Response.connected(0, "User0"));
        // The following assertEquals tests whether the changeNickname method returns
        // the expected ResponseSet.
        assertEquals(expected, model.registerUser(0), "User0 registered");

        // Alternatively, we can also instantiate a new ResponseSet using
        // the default constructor.
        ResponseSet expected2 = new ResponseSet();
        // Then, add the response to the ResponseSet using addMessage
        expected2.addMessage(Response.connected(0, "User0"));
        // It should create the same expected ResponseSet as using
        // singleMessage.
        assertEquals(expected, expected2, "Different ways of creating a ResponseSet");
        // Note that you don't have to use both ways to create
        // the expected ResponseSet in each test.
        // This test is just included to show you the two different
        // ways of constructing a ResponseSet. Choose whatever works best
        // in each scenario!

        // test: does the collection contain only one element?
        Collection<String> registeredUsers = model.getRegisteredUsers();
        assertEquals(1, registeredUsers.size(), "Num. registered users");

        // test: does the collection contain the nickname "User0"
        assertTrue(registeredUsers.contains("User0"), "User0 is registered");
    }

    /*
     * Your TAs will be manually grading the tests that you write below this
     * comment block. Don't forget to test the public methods you have added to
     * your ServerModel class, as well as the behavior of the server in
     * different scenarios.
     * You might find it helpful to take a look at the tests we have already
     * provided you with in Task4Test, Task3Test, and Task5Test.
     */

}
